package fpcs.core;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

/**
 *
 * @author Ignas Kukenys
 *
 */
public class ClassifierSet implements Iterable<Classifier> {

    protected ArrayList<Classifier> classifiers = new ArrayList<Classifier>();
    protected int numerositySum;

    public ClassifierSet() {
    }

    public ClassifierSet(List<Classifier> entries) {
        numerositySum = 0;
        for (Classifier c : entries) {
            numerositySum += c.getNumerosity();
            classifiers.add(c);
        }
    }

    public int size() {
        return classifiers.size();
    }

    public Classifier get(int i) {
        return classifiers.get(i);
    }

    public int getNumerositySum() {
        return numerositySum;
    }

    public int indexOf(Classifier classifier) {
        return classifiers.indexOf(classifier);
    }

    public boolean contains(Classifier classifier) {
        return indexOf(classifier) > -1;
    }

    public boolean isActionCovered(int action) {
        for (Classifier clsf : classifiers) {
            if (clsf.getAction() == action) {
                return true;
            }
        }
        return false;
    }

    public void addSubsume(Classifier classifier) {
        Classifier existing = findIdenticalClassifier(classifier);
        if (existing != null) {
            existing.addNumerosity(classifier.getNumerosity());
            addNumerositySum(classifier.getNumerosity());
        } else {
            add(classifier);
        }
    }

    public void addAll(ClassifierSet other) {
        for (Classifier c : other) {
            this.add(c);
        }
    }

    public void add(Classifier classifier) {
        classifiers.add(classifier);
        numerositySum += classifier.getNumerosity();
    }

    public void remove(Classifier classifier) {
        if (classifiers.remove(classifier)) {
            numerositySum -= classifier.getNumerosity();
        }
    }

    public void removeEmpty(Classifier classifier) {
        if (classifier.getNumerosity() == 0) {
            classifiers.remove(classifier);
        }
    }

    public void addNumerositySum(int numerosity) {
        numerositySum += numerosity;
    }

//    protected double getPredictionSum() {
//        double sum = 0.0;
//
//        for (Classifier classifier : classifiers) {
//            sum += classifier.getPrediction() * classifier.getNumerosity();
//        }
//        return sum;
//    }

    public double getFitnessSum() {
        double sum = 0.0;

        for (Classifier classifier : classifiers) {
            sum += classifier.getFitness();
        }
        return sum;
    }

    public double getTimeStampSum() {
        double sum = 0.0;

        for (Classifier classifier : classifiers) {
            sum += classifier.getTimeStamp() * classifier.getNumerosity();
        }
        return sum;
    }

    public double getTimeStampAverage() {
        return getTimeStampSum() / numerositySum;
    }

    public Classifier findIdenticalClassifier(Classifier target) {
        for (Classifier classifier : classifiers) {
            if (target.isIdentical(classifier)) {
                return classifier;
            }
        }
        return null;
    }

    public void setTimeStamps(int timestamp) {
        for (Classifier classifier : classifiers) {
            classifier.setTimeStamp(timestamp);
        }
    }

    public double getDeletionVoteSum(double meanFitness) {
        double sum = 0.0;
        for (Classifier classifier : classifiers) {
            sum += classifier.getDeletionVote(meanFitness);
        }
        return sum;
    }

    /**
     * @return tournament selected classifier, according to numerosities
     */
    public Classifier selectForBreedingT() {

        Classifier selection = null;

        while (selection == null) {

            double bestFitness = 0.0;

            for (int i = 0; i < classifiers.size(); i++) {

                int numerosity = classifiers.get(i).getNumerosity();

                if (classifiers.get(i).getFitness() / numerosity > bestFitness) {

                    for (int j = 0; j < numerosity; j++) {
                        if (RandomUtils.random() < LearningParams.TOURNAMENT_FRACTION_TAU) {
                            selection = classifiers.get(i);
                            bestFitness = classifiers.get(i).getFitness() / numerosity;
                            break;
                        }
                    }
                }
            }
        }
        return selection;
    }

    public Classifier selectForBreedingRW(double fitSum) {
        double choiceP = RandomUtils.random() * fitSum;
        int i = 0;
        double sum = classifiers.get(i).getFitness();
        while (choiceP > sum) {
            i++;
            sum += classifiers.get(i).getFitness();
        }

        return classifiers.get(i);
    }

    public Classifier selectForDeletionRW() {
        double fitnessSum = getFitnessSum();
        double meanFitness = fitnessSum / (double) numerositySum;
        double voteSum = getDeletionVoteSum(meanFitness);
        double choicePoint = voteSum * RandomUtils.random();
        double sum = 0.0;

        //System.out.println(String.format("Select for deletion: fitSum %.2f meanFitness %.2f, votesum %.2f choicePoint %.2f", fitnessSum, meanFitness, voteSum, choicePoint));

        for (int i = 0; i < classifiers.size(); i++) {
            sum += classifiers.get(i).getDeletionVote(meanFitness);
            if (sum > choicePoint) {
                return classifiers.get(i);
            }
        }
        return null;
    }

    public boolean subsumeClassifier(Classifier cl) {

        ArrayList<Classifier> choices = new ArrayList<Classifier>();
        for (Classifier classifier : classifiers) {
            if (classifier.subsumes(cl)) {
                choices.add(classifier);
            }
        }

        if (choices.size() > 0) {
            int choice = (int) (RandomUtils.random() * choices.size());
            choices.get(choice).addNumerosity(+1);
            addNumerositySum(+1);
            return true;
        } else {
            //If no subsumer was found, add the classifier to the population
            //addSubsume(cl);
            return false;
        }
    }

  

//    protected void updateFitnessSet() {
//        double accuracySum = 0.0;
//        double[] accuracies = new double[classifiers.size()];
//
//        //First, calculate the accuracies of the classifier and the accuracy sums
//        for (int i = 0; i < classifiers.size(); i++) {
//            accuracies[i] = classifiers.get(i).getAccuracy();
//            accuracySum += accuracies[i] * classifiers.get(i).getNumerosity();
//        }
//
//        //Next, update the fitnesses accordingly
//        for (int i = 0; i < classifiers.size(); i++) {
//            classifiers.get(i).updateFitness(accuracySum, accuracies[i]);
//        }
//    }

    public void reduceToSize(int size) {
        while (numerositySum > size) {
            Classifier toRemove = selectForDeletionRW();
            if (toRemove != null) {
                toRemove.addNumerosity(-1);
                addNumerositySum(-1);
                removeEmpty(toRemove);
            }
        }
    }

    @Override
    public Iterator<Classifier> iterator() {
        return classifiers.iterator();
    }

    @Override
    public String toString() {
        return String.format("Set(%d | %d)", this.classifiers.size(), this.numerositySum);
    }

    public void printAllByFitness() {
        printAllByFitness(Integer.MAX_VALUE);
    }

    public void printAllByFitness(int topN) {

        List<Classifier> byFitness = makeClassifierListByFitness();

        for (int i = 0; i < byFitness.size() && i < topN; i++) {
            Classifier xc = byFitness.get(byFitness.size() - i - 1);
            System.out.println(xc);
        }
    }

    public List<Classifier> makeClassifierListByFitness() {
        List<Classifier> result = new ArrayList<Classifier>();
        result.addAll(classifiers);

        Collections.sort(result, new Comparator<Classifier>() {

            @Override
            public int compare(Classifier o1, Classifier o2) {
                if (o1.getFitness() < o2.getFitness()) {
                    return -1;
                } else if (o1.getFitness() > o2.getFitness()) {
                    return 1;
                } else {
                    return 0;
                }
            }
        });

        return result;
    }

    public List<Classifier> makeClassifierListByAccuracy() {
        List<Classifier> result = new ArrayList<Classifier>();
        result.addAll(classifiers);

        Collections.sort(result, new Comparator<Classifier>() {

            @Override
            public int compare(Classifier o1, Classifier o2) {
                if (o1.getAccuracy() < o2.getAccuracy()) {
                    return -1;
                } else if (o1.getAccuracy() > o2.getAccuracy()) {
                    return 1;
                } else {
                    return 0;
                }
                //return (o1.getAccuracy() - o2.getAccuracy()) < 0?-1:1;
            }
        });

        return result;
    }

//    public List<Classifier> makeClassifierListByPrediction() {
//        List<Classifier> result = new ArrayList<Classifier>();
//        result.addAll(classifiers);
//
//        Collections.sort(result, new Comparator<Classifier>() {
//
//            @Override
//            public int compare(Classifier o1, Classifier o2) {
//                if (o1.getPrediction() < o2.getPrediction()) {
//                    return -1;
//                } else if (o1.getPrediction() > o2.getPrediction()) {
//                    return 1;
//                } else {
//                    return 0;
//                }
//            }
//        });
//
//        return result;
//    }

    public void updateCorrectTrack() {
        for (Classifier clsf : classifiers) {
            clsf.increaseCorrectTrack();
        }
    }

    public void updateSet() {
        for (Classifier clsf : classifiers) {
            clsf.increaseExperience();
            clsf.updateAccuracy();
            clsf.updateActionSetSize(this.numerositySum);
            clsf.updateFitness();

        }

    }


    public void removeClassifiers(){
        this.classifiers.removeAll(classifiers);

    }


}
