package fpcs.image;

import fpcs.core.Classifier;
import fpcs.core.ClassifierSet;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Scanner;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


/**
 * Saves classifiers using @@Classifier.toString
 * Loads classifiers from text files as output by @@Classifier.toString
 *
 * @author Ignas Kukenys
 *
 */

public class ClassifierIO {

    private static Pattern classifierPattern = Pattern.compile("^X\\s*(\\d+) N\\s*(\\d+) F([-+]?[0-9]*\\.?[0-9]+) P\\s*([-+]?[0-9]*\\.?[0-9]+) A([-+]?[0-9]*\\.?[0-9]+) .* -> (\\d+)$");
    private static Pattern featurePattern = Pattern.compile("(HAAR.+?): (\\d+?),(\\d+?)-(\\d+?)x(\\d+?) in \\(\\s*([-+]?[0-9]*\\.?[0-9]+),\\s*([-+]?[0-9]*\\.?[0-9]+)\\)");
    private static double MIN_FITNESS = 0.000001;
    /**
     * Loads classifiers from text files as output by @@Classifier.toString
     *
     * Warning: not stable if toString methods of Classifiers or Conditions change, uses regular expressions to parse output
     *
     * @param filename containing classifiers
     * @return ClassifierSet
     */
    public static ClassifierSet loadClassifiers(String filename) {
        ClassifierSet result = new ClassifierSet();

        try {
            Scanner lines = new Scanner(new File(filename));

            while(lines.hasNextLine()) {

                String line = lines.nextLine();
                Matcher classifierMatcher = classifierPattern.matcher(line);

                if (!line.isEmpty() && classifierMatcher.matches()) {

                    int experience = Integer.parseInt(classifierMatcher.group(1));
                    int numerosity = Integer.parseInt(classifierMatcher.group(2));
                    double fitness = Double.parseDouble(classifierMatcher.group(3));
                    if (fitness == 0) {
                        fitness = MIN_FITNESS;
                    }
                    double prediction = Double.parseDouble(classifierMatcher.group(4));
                    //double accuracy = Double.parseDouble(macho.group(5));
                    int action = Integer.parseInt(classifierMatcher.group(6));

                    Classifier cl = new ImageClassifier();
                    cl.addNumerosity(numerosity);
                    cl.setExperience(experience);
                    cl.setAction(action);
//                    cl.setPrediction(prediction);
                    cl.setFitness(fitness);

                    FeatureCondition fc = new FeatureCondition(new ArrayList<HaarRangeFeature>());
                    cl.setCondition(fc);

                    Matcher featureMatcher = featurePattern.matcher(line);

                    while(featureMatcher.find()) {
                        HaarFeatureType type = HaarFeatureType.valueOf(featureMatcher.group(1));
                        int x = Integer.parseInt(featureMatcher.group(2));
                        int y = Integer.parseInt(featureMatcher.group(3));
                        int hs = Integer.parseInt(featureMatcher.group(4)) / type.width;
                        int vs = Integer.parseInt(featureMatcher.group(5)) / type.height;
                        double lo = Double.parseDouble(featureMatcher.group(6));
                        double hi = Double.parseDouble(featureMatcher.group(7));
                        fc.getHaarRangeFeatures().add(new HaarRangeFeature(type, x, y, hs, vs, lo, hi));
                    }

                    result.add(cl);
                }

            }

        } catch (FileNotFoundException ex) {
            throw new RuntimeException("File not found: " + filename, ex);
        } finally {
            return result;
        }

    }

    public static void saveClassifiers(ClassifierSet classifiers, String filename) {

        PrintWriter w;
        try {
            w = new PrintWriter(new BufferedWriter(new FileWriter(filename)));

            for (Classifier c: classifiers) {
                w.println(c);
            }

            w.flush();
            w.close();

        } catch (IOException ex) {
            throw new RuntimeException("Unable to write to: " + filename, ex);
        }

    }



}
