package fpcs.image;

import java.util.ArrayList;
import java.util.List;
import fpcs.core.Condition;
import fpcs.core.LearningParams;
import fpcs.core.RandomUtils;
import fpcs.core.State;

/**
 *
 * @author Ignas Kukenys
 *
 */
public class FeatureCondition extends Condition {
    
    public static FeatureCondition generateForState(ImageState situation) {
        
        ArrayList<HaarRangeFeature> list = new ArrayList<HaarRangeFeature>();
        for (int i = 0; i < (int)(RandomUtils.random() * LearningParams.MESSY_MAX_FEATURES) + 1; i++) {
            list.add(HaarRangeFeature.randomMatching(situation));
        }

        return new FeatureCondition(list);
    }

    private ArrayList<HaarRangeFeature> features = new ArrayList<HaarRangeFeature>();


    public FeatureCondition(List<HaarRangeFeature> features) {
        this.features.addAll(features);
    }


    public List<HaarRangeFeature> getHaarRangeFeatures() {
        return features;
    }

    @Override
    public boolean matches(State state) {

        ImageState istate = (ImageState)state;
        for (HaarRangeFeature f: features) {
            boolean t=!f.presentIn(istate);
            if (!f.presentIn(istate)) {               
                return false;
            }
        }
        return true;
    }

    @Override
    public boolean isMoreGeneral(Condition cond) {

        //System.out.println("Proper isMoreGeneral used");

        FeatureCondition fcond = (FeatureCondition)cond;

        if (this.features.size() <= fcond.features.size()) {

            for (HaarRangeFeature f1: fcond.features) {
                boolean foundMoreGeneral = false;
                for(HaarRangeFeature f2: this.features) {
                    if (f2.isMoreGeneralThan(f1)) {
                        foundMoreGeneral = true;
                        break;
                    }
                }

                if (!foundMoreGeneral) {
                    return false;
                }

            }

            return true; // this is shorter and covers all features in cond.

        }
        return false;

        //return this.isIdentical(fcond);
    }

    @Override
    public Condition clone() {
        List<HaarRangeFeature> list = new ArrayList<HaarRangeFeature>();

        for (HaarRangeFeature feature : this.features) {
            list.add(feature.clone());
        }

        return new FeatureCondition(list);
    }

    @Override
    public boolean isIdentical(Condition cond) {

        //throw new RuntimeException("Not supported!!");

        if (cond instanceof FeatureCondition) {
            FeatureCondition fcond = (FeatureCondition)cond;
            // TODO: ignore order?
            if (fcond.features.size() == this.features.size()) {
                for (int i = 0; i < this.features.size(); i++) {
                    if (!fcond.features.get(i).isIdentical(this.features.get(i))) {
                        return false;
                    }
                }
                return true;
            }
        }
        return false;
    }

    public boolean mutate(ImageState state) {
        boolean valid = true;

        for (HaarRangeFeature f : features) {
            f.mutate(state);
        }

        if (RandomUtils.random() < LearningParams.MESSY_PROB_SHRINK && features.size() > 1) {
            features.remove((int)(RandomUtils.random() * features.size()));
        }

        if (features.isEmpty() || (RandomUtils.random() < LearningParams.MESSY_PROB_EXPAND && features.size() < LearningParams.MESSY_MAX_FEATURES)) {
            features.add((int)(RandomUtils.random() * (features.size() + 1)), HaarRangeFeature.randomMatching(state));
        }

        for (HaarRangeFeature f : features) {
            valid = valid && f.isValid(state);
        }

        return valid;
    }

    public boolean crossover(Condition condition) {
        
        //boolean valid = true;

        List<HaarRangeFeature> other = ((FeatureCondition) condition).getHaarRangeFeatures();
        List<HaarRangeFeature> longer, shorter;
        if (other.size() > features.size()) {
            longer = other;
            shorter = features;
        } else {
            longer = features;
            shorter = other;
        }


        List<HaarRangeFeature> tmp1 = new ArrayList<HaarRangeFeature>();
        List<HaarRangeFeature> tmp2 = new ArrayList<HaarRangeFeature>();

        for (HaarRangeFeature feature: longer) {
            if (RandomUtils.random() < LearningParams.PROB_CROSSOVER_UNIFORM) {
                tmp1.add(feature);
            }
        }

        for (HaarRangeFeature feature: shorter) {
            if (RandomUtils.random() < LearningParams.PROB_CROSSOVER_UNIFORM) {
                tmp2.add(feature);
            }
        }

        longer.removeAll(tmp1);
        shorter.removeAll(tmp2);
        longer.addAll(tmp2);
        shorter.addAll(tmp1);

        return true;
    }

    @Override
    public String toString() {

        StringBuilder sb = new StringBuilder(100);


        for (HaarRangeFeature f: features) {
            if (sb.length() > 0) {
                sb.append(" && ");
            }
            sb.append(f.toString());
        }

        return sb.toString();
    }

}
