package fpcs.image;

import fpcs.core.RandomUtils;

/**
 *
 * @author Ignas Kukenys
 *
 */
public enum HaarFeatureType {

    HAAR1(1, 1) {

        @Override
        public double evaluate(ImageState state, int x, int y, int hScale, int vScale) {
            double t1 = state.integral(x - 1, y - 1);
            double t2 = state.integral(x + hScale - 1, y - 1);

            double b1 = state.integral(x - 1, y + vScale - 1);
            double b2 = state.integral(x + hScale - 1, y + vScale - 1);


            return b2 - t2 - b1 + t1;
        }
    },
    HAAR2HOR(2, 1) {

        @Override
        public double evaluate(ImageState state, int x, int y, int hScale, int vScale) {
            double t1 = state.integral(x - 1, y - 1);
            double t2 = state.integral(x + hScale - 1, y - 1);
            double t3 = state.integral(x + (hScale * 2) - 1, y - 1);

            double b1 = state.integral(x - 1, y + vScale - 1);
            double b2 = state.integral(x + hScale - 1, y + vScale - 1);
            double b3 = state.integral(x + (hScale * 2) - 1, y + vScale - 1);

            return b3 - t3 - b2 + t2 - b2 + t2 + b1 - t1;
        }
    },
    HAAR2VER(1, 2) {

        /*
         *  -
         *  +
         *
         */
        @Override
        public double evaluate(ImageState state, int x, int y, int hScale, int vScale) {
            double l1 = state.integral(x - 1, y - 1);
            double r1 = state.integral(x + hScale - 1, y - 1);
            double l2 = state.integral(x - 1, y + vScale - 1);
            double r2 = state.integral(x + hScale - 1, y + vScale - 1);
            double l3 = state.integral(x - 1, y + (vScale * 2) - 1);
            double r3 = state.integral(x + hScale - 1, y + (vScale * 2) - 1);

            return (r3 - r2 - l3 + l2) - (r2 - r1 - l2 + l1);
        }
    },
    HAAR3HOR(3, 1) {

        /*
         *
         *    - + -
         *
         */
        @Override
        public double evaluate(ImageState state, int x, int y, int hScale, int vScale) {
            double t1 = state.integral(x - 1, y - 1);
            double t2 = state.integral(x + hScale - 1, y - 1);
            double t3 = state.integral(x + (hScale * 2) - 1, y - 1);
            double t4 = state.integral(x + (hScale * 3) - 1, y - 1);

            double b1 = state.integral(x - 1, y + vScale - 1);
            double b2 = state.integral(x + hScale - 1, y + vScale - 1);
            double b3 = state.integral(x + (hScale * 2) - 1, y + vScale - 1);
            double b4 = state.integral(x + (hScale * 3) - 1, y + vScale - 1);

            return (b3 - t3 - b2 + t2) - (b2 - t2 - b1 + t1) - (b4 - t4 - b3 + t3);
        }
    },
    HAAR3VER(1, 3) {

        /*
         *    -
         *    +
         *    -
         *
         */
        @Override
        public double evaluate(ImageState state, int x, int y, int hScale, int vScale) {
            double l1 = state.integral(x - 1, y - 1);
            double r1 = state.integral(x + hScale - 1, y - 1);
            double l2 = state.integral(x - 1, y + vScale - 1);
            double r2 = state.integral(x + hScale - 1, y + vScale - 1);
            double l3 = state.integral(x - 1, y + (vScale * 2) - 1);
            double r3 = state.integral(x + hScale - 1, y + (vScale * 2) - 1);
            double l4 = state.integral(x - 1, y + (vScale * 3) - 1);
            double r4 = state.integral(x + hScale - 1, y + (vScale * 3) - 1);

            return (r3 - r2 - l3 + l2) - (r2 - r1 - l2 + l1) - (r4 - r3 - l4 + l3);
        }
    },
    HAAR4(2, 2) {
        /*
         *     - +
         *     + -
         */

        @Override
        public double evaluate(ImageState state, int x, int y, int hScale, int vScale) {
            double t1 = state.integral(x - 1, y - 1);
            double t2 = state.integral(x + hScale - 1, y - 1);
            double t3 = state.integral(x + (hScale * 2) - 1, y - 1);

            double m1 = state.integral(x - 1, y + vScale - 1);
            double m2 = state.integral(x + hScale - 1, y + vScale - 1);
            double m3 = state.integral(x + (hScale * 2) - 1, y + vScale - 1);

            double b1 = state.integral(x - 1, y + (vScale * 2) - 1);
            double b2 = state.integral(x + hScale - 1, y + (vScale * 2) - 1);
            double b3 = state.integral(x + (hScale * 2) - 1, y + (vScale * 2) - 1);

            return (m3 - t3 - m2 + t2) + (b2 - m2 - b1 + m1) - (m2 - t2 - m1 + t1) - (b3 - m3 - b2 + m2);

        }
    };
    public final int width;
    public final int height;

    private static final HaarFeatureType[] values = HaarFeatureType.values();

    public static HaarFeatureType randomType() {
        return values[(int) (RandomUtils.random() * values.length)];

    }

    private HaarFeatureType(int w, int h) {
        width = w;
        height = h;
    }

    public abstract double evaluate(ImageState state, int x, int y, int hScale, int vScale);
}
