package fpcs.image;

import fpcs.core.LearningParams;
import fpcs.core.RandomUtils;

/**
 *
 * @author Ignas Kukenys
 *
 */
public class HaarRangeFeature {

    

    public static HaarRangeFeature randomMatching(ImageState state) {
        return new HaarRangeFeature(state);
    }
    
    private HaarFeatureType type;
    private int x;
    private int y;
    private int hScale;
    private int vScale;
    private double tLow, tHigh;

    private HaarRangeFeature(HaarRangeFeature copy) {
        this.type = copy.type;
        this.x = copy.x;
        this.y = copy.y;
        this.hScale = copy.hScale;
        this.vScale = copy.vScale;
        this.tLow = copy.tLow;
        this.tHigh = copy.tHigh;
    }

    private HaarRangeFeature(ImageState state) {

        this.type = HaarFeatureType.randomType();

        int maxhScale = state.width / type.width;
        int maxvScale = state.height / type.height;
        this.hScale = RandomUtils.getRandomizer().nextInt(maxhScale - 1) + 1 + 1;
        this.vScale = RandomUtils.getRandomizer().nextInt(maxvScale - 1) + 1 + 1;

        int maxx = state.width - type.width * hScale;
        int maxy = state.height - type.height * vScale;

        // TODO: parameterize position placing

        //this.x = random.nextInt(maxx + 1);
        //this.y = random.nextInt(maxy + 1);

        double max_value = 0;
        for (int j = 0; j <= maxy; j++) {
            for (int i = 0; i <= maxx; i++) {
                double t = Math.abs(this.type.evaluate(state, i, j, hScale, vScale));
                if (t > max_value) {
                    this.x = i;
                    this.y = j;
                    max_value = t;
                }
            }
        }

        int maxT = hScale * vScale * 255;

        double value = evaluate(state);

        
        this.tLow = (int) (value - Math.abs(value / 2));
        this.tHigh = (int) (value + Math.abs(value / 2));

        assert (x >= 0) && (x <= state.width - type.width * hScale)
                && (y >= 0) && (y <= state.height - type.height * vScale)
                && (hScale > 0) && (vScale > 0) : "HaarFeature construction for image failed";


    }

    public HaarRangeFeature(HaarFeatureType type, int x, int y, int hScale, int vScale, double low, double high) {
        this.type = type;
        this.x = x;
        this.y = y;
        this.hScale = hScale;
        this.vScale = vScale;
        this.tLow = low;
        this.tHigh = high;
    }

    public boolean presentIn(ImageState state) {
        double value = evaluate(state);
        return (value >= tLow) && (value <= tHigh);
    }

    private double evaluate(ImageState state) {
        return this.type.evaluate(state, x, y, hScale, vScale);
    }

    public boolean isValid(ImageState state) {
        return (x >= 0) && (x <= state.width - type.width * hScale)
                && (y >= 0) && (y <= state.height - type.height * vScale)
                && (hScale > 0) && (vScale > 0);
    }

    public boolean mutate(ImageState state) {

        if (RandomUtils.random() < LearningParams.PROB_MUTATION_UNIFORM) {
            x += RandomUtils.getRandomizer().nextInt(2) * 2 - 1;
        }

        if (RandomUtils.random() < LearningParams.PROB_MUTATION_UNIFORM) {
            y += RandomUtils.getRandomizer().nextInt(2) * 2 - 1;
        }

        if (RandomUtils.random() < LearningParams.PROB_MUTATION_UNIFORM) {
            hScale += RandomUtils.getRandomizer().nextInt(2) * 2 - 1; // +/-1
        }

        if (RandomUtils.random() < LearningParams.PROB_MUTATION_UNIFORM) {
            vScale += RandomUtils.getRandomizer().nextInt(2) * 2 - 1; // +/-1
        }

        if (isValid(state)) {

            int maxT = hScale * vScale * 255;
            double value = evaluate(state);

            if (RandomUtils.random() < LearningParams.PROB_MUTATION_UNIFORM) {
                this.tLow += RandomUtils.between(-maxT / 255, maxT / 255);
            }

            if (RandomUtils.random() < LearningParams.PROB_MUTATION_UNIFORM) {
                this.tHigh += RandomUtils.between(-maxT / 255, maxT / 255);
            }

            if (this.tLow > value) {
                this.tLow = value - 1.0;
            }

            if (this.tHigh < value) {
                this.tHigh = value + 1.0;
            }


            return true;
        } else {
            return false;
        }

    }
    /*
    public boolean crossover(HaarRangeFeature mate) {
    if (RandomUtils.random() < LearningParams.PROB_CROSSOVER_UNIFORM) {
    int tmp = this.x;
    this.x = mate.x;
    mate.x = tmp;
    }

    if (RandomUtils.random() < LearningParams.PROB_CROSSOVER_UNIFORM) {
    int tmp = this.y;
    this.y = mate.y;
    mate.y = tmp;
    }

    if (RandomUtils.random() < LearningParams.PROB_CROSSOVER_UNIFORM) {
    int tmp = this.hScale;
    this.hScale = mate.hScale;
    mate.hScale = tmp;
    }

    if (RandomUtils.random() < LearningParams.PROB_CROSSOVER_UNIFORM) {
    int tmp = this.vScale;
    this.vScale = mate.vScale;
    mate.vScale = tmp;
    }

    if (RandomUtils.random() < LearningParams.PROB_CROSSOVER_UNIFORM) {
    int tmp = this.parity;
    this.parity = mate.parity;
    mate.parity = tmp;
    }

    if (RandomUtils.random() < LearningParams.PROB_CROSSOVER_UNIFORM) {
    double tmp = this.tLow;
    this.tLow = mate.tLow;
    mate.tLow = tmp;
    }

    if (RandomUtils.random() < LearningParams.PROB_CROSSOVER_UNIFORM) {
    int tmp = this.parity;
    this.parity = mate.parity;
    mate.parity = tmp;
    }


    return true;

    }*/

    @Override
    public HaarRangeFeature clone() {
        return new HaarRangeFeature(this);
    }

    @Override
    public String toString() {
        return String.format("%s: %d,%d-%dx%d in (%7.2f,%7.2f)",
                type,
                x, y,
                type.width * hScale,
                type.height * vScale,
                tLow, tHigh);
    }

    public boolean isIdentical(HaarRangeFeature feature) {

        return (this.type == feature.type)
                && (this.x == feature.x)
                && (this.y == feature.y)
                && (this.hScale == feature.hScale)
                && (this.vScale == feature.vScale)
                && (this.tLow == feature.tLow)
                && (this.tHigh == feature.tHigh);
    }

    public boolean isMoreGeneralThan(HaarRangeFeature feature) {
        return (this.type == feature.type)
                && (this.x == feature.x)
                && (this.y == feature.y)
                && (this.hScale == feature.hScale)
                && (this.vScale == feature.vScale)
                && (this.tLow <= feature.tLow)
                && (this.tHigh >= feature.tHigh);
    }
}
