package fpcs.symbolic;

import fpcs.core.Classifier;
import fpcs.core.LearningParams;
import fpcs.core.RandomUtils;
import fpcs.core.State;

public class SymbolicClassifier extends Classifier {

    /**
     * The don't care symbol (normally '#')
     */
    public static char SYMBOL_DONTCARE = '#';

    /**
     * Constructs a classifier with matching condition and specified action.
     *
     * @param setSize The size of the current set which the new classifier matches.
     * @param time  The actual number of instances the XCS learned from so far.
     * @param situation The current problem instance/perception.
     * @param act The action of the new classifier.
     */
    public SymbolicClassifier(State situation, int act, double setSize, int timestamp) {
        super(setSize, timestamp);
        createMatchingCondition(situation);
        action = act;
    }

    /**
     * Construct matching classifier with random action.
     *
     * @param setSize The size of the current set which the new classifier matches.
     * @param time The actual number of instances the XCS learned from so far.
     * @param numberOfActions The number of different actions to chose from 
     * (This should be set to the number of actions possible in the problem).
     * @param situation The current problem instance/perception.


    private SymbolicClassifier(double setSize, int time, int numberOfActions, State situation) {
        super(setSize, time);
        createMatchingCondition(situation);
        createRandomAction(numberOfActions);
    }
     *
     */

    /**
     * Construct a classifier with random condition and random action.
     *
     * @param setSize The size of the current set which the new classifier matches.
     * @param time  The actual number of instances the XCS learned from so far.
     * @param condLength The length of the condition of the new classifier.
     * @param numberOfActions The number of different actions to chose from 
     
    private SymbolicClassifier(double setSize, int time, int condLength, int numberOfActions) {
        super(setSize, time);
        createRandomCondition(condLength);
        createRandomAction(numberOfActions);
        
    }
     *
     */

    /**
     * Constructs an identical SymbolicClassifier.
     * However, the experience of the copy is set to 0 and the numerosity is set to 1 since this is indeed 
     * a new individual in a population.
     *
     * @param clOld The to be copied classifier.
     */
    private SymbolicClassifier(SymbolicClassifier clOld) {

        if(clOld.numerosity == 0) {
            throw new RuntimeException("Cannot clone classifier, numerosity = 0");
        }

        condition = clOld.getCondition().clone();
        action = clOld.getAction();
//        this.prediction = clOld.getPrediction();
//        this.predictionError = clOld.getPredictionError();
        // Here we should divide the fitness by the numerosity to get a accurate value for the new one!
        this.fitness = clOld.fitness / clOld.numerosity;
        this.numerosity = 1;
        this.experience = 0;
        this.actionSetSize = clOld.actionSetSize;
        this.timeStamp = clOld.timeStamp;
    }

    /**
     * Creates a condition randomly considering the constant <code>PROB_DONTCARE<\code>.
     *
     * @see LearningParams#PROB_DONTCARE
     
    private void createRandomCondition(int condLength) {
        char condArray[] = new char[condLength];
        for (int i = 0; i < condLength; i++) {
            if (RandomUtils.random() < LearningParams.PROB_DONTCARE) {
                condArray[i] = LearningParams.SYMBOL_DONTCARE;
            } else if (RandomUtils.random() < 0.5) {
                condArray[i] = '0';
            } else {
                condArray[i] = '1';
            }
        }
        condition = new SymbolicCondition(condArray);
    }
     *
     */

    /**
     * Creates a matching condition considering the constant <code>PROB_DONTCARE<\code>.
     *
     * @see LearningParams#PROB_DONTCARE
     */
    private void createMatchingCondition(State state) {
        String stateString = ((SymbolicState)state).getStateAsString();
        int condLength = stateString.length();
        char condArray[] = new char[condLength];

        for (int i = 0; i < condLength; i++) {
            if (RandomUtils.random() < LearningParams.PROB_DONTCARE) {
                condArray[i] = SymbolicClassifier.SYMBOL_DONTCARE;
            } else {
                condArray[i] = stateString.charAt(i);
            }
        }
        condition = new SymbolicCondition(condArray);
    }

    /**
     * Creates a random action.
     *
     * @param numberOfActions The number of actions to chose from.
     
    private void createRandomAction(int numberOfActions) {
        action = (int) (RandomUtils.random() * numberOfActions);
    }
     *
     */


    /**
     * Applies two point crossover and returns if the classifiers changed.
     *
     * @see LearningParams#PROB_CROSSOVER_CHI
     * @param cl The second classifier for the crossover application.
     */
    public boolean applyCrossover(Classifier cl) {
        boolean changed = false;
        String condString = getCondition().getConditionAsString();
        String otherCondString = ((SymbolicClassifier)cl).getCondition().getConditionAsString();
        if (RandomUtils.random() < LearningParams.PROB_CROSSOVER_CHI) {
            int length = condString.length();
            int sep1 = (int) (RandomUtils.random() * (length));
            int sep2 = (int) (RandomUtils.random() * (length)) + 1;
            if (sep1 > sep2) {
                int help = sep1;
                sep1 = sep2;
                sep2 = help;
            } else if (sep1 == sep2) {
                sep2++;
            }
            char[] cond1 = condString.toCharArray();
            char[] cond2 = otherCondString.toCharArray();
            for (int i = sep1; i < sep2; i++) {
                if (cond1[i] != cond2[i]) {
                    changed = true;
                    char help = cond1[i];
                    cond1[i] = cond2[i];
                    cond2[i] = help;
                }
            }
            if (changed) {
                condition = new SymbolicCondition(cond1);
                cl.setCondition(new SymbolicCondition(cond2));
            }
        }
        return changed;
    }

    /**
     * @param state The current situation/problem instance
     * @param numberOfActions The maximal number of actions possible in the environment.
     * @return true if mutation produced a valid classifier
     */
    @Override
    public boolean applyMutation(State state, int numberOfActions) {
        mutateAction(numberOfActions);
        return mutateCondition(state);
    }

    /**
     * Mutates the condition of the classifier. If one allele is mutated depends on the constant PROB_MUTATION_MU.
     * This mutation is a niche mutation. It assures that the resulting classifier
     * still matches the current situation.
     *
     * @see LearningParams#PROB_MUTATION_MU
     * @param state The current situation/problem instance.
     */
    private boolean mutateCondition(State state) {
        //boolean changed = false;
        String condString = getCondition().getConditionAsString();
        int condLength = condString.length();
        char[] cond = condString.toCharArray();
        char[] stateC = ((SymbolicState)state).getStateAsString().toCharArray();

        for (int i = 0; i < condLength; i++) {
            if (RandomUtils.random() < LearningParams.PROB_MUTATION_MU) {
                //changed = true;
                if (cond[i] == SymbolicClassifier.SYMBOL_DONTCARE) {
                    cond[i] = stateC[i];
                } else {
                    cond[i] = SymbolicClassifier.SYMBOL_DONTCARE;
                }
            }
        }
        condition = new SymbolicCondition(cond);
        return true; // meaning the mutation produced a valid offspring
    }


    @Override
    public boolean isIdentical(Classifier classifier) {

        if (classifier instanceof SymbolicClassifier) {
            SymbolicClassifier cl = (SymbolicClassifier)classifier;
            if (cl.getCondition().isIdentical(condition)) {
                if (cl.getAction() == action) {
                    return true;
                }
            }
        }
        return false;
    }


    @Override
    public SymbolicCondition getCondition(){
        return (SymbolicCondition)condition;
    }


    @Override
    public SymbolicClassifier offspring() {
        return new SymbolicClassifier(this);
    }

}
