package fpcs.utils;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.image.BandedSampleModel;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferDouble;
import java.awt.image.Raster;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import javax.imageio.ImageIO;

public class ImageUtils {

    private static int THUMBNAIL_SPACING = 5;

    public static BufferedImage renderThumbnails(ArrayList<double[]> list, Dimension d) {

        int w = (int) Math.sqrt(list.size());
        int h = list.size() / w + ((list.size() % w == 0) ? 0 : 1);

        int ww = d.width + THUMBNAIL_SPACING;
        int hh = d.height + THUMBNAIL_SPACING;

        BufferedImage im = new BufferedImage(w * ww + THUMBNAIL_SPACING, h * hh + THUMBNAIL_SPACING, BufferedImage.TYPE_BYTE_GRAY);

        Graphics g = im.getGraphics();

        int i = 0;

        for (int row = 0; row < h; row++) {
            int col = 0;
            while (col < w && i < list.size()) {
                BufferedImage patch = ImageUtils.array2image(list.get(i), d);
                g.drawImage(patch, 5 + col * ww, 5 + row * hh, null);
                i++;
                col++;
            }
        }

        g.dispose();

        return im;
    }

    public static BufferedImage renderThumbnails(double[][] array, Dimension d) {
        ArrayList<double[]> list = new ArrayList<double[]>(array.length);
        list.addAll(Arrays.asList(array));
        return renderThumbnails(list, d);
    }

    public static BufferedImage image2grayscale(BufferedImage image, boolean forceNewImage) {

        if (image == null) {
            return null;
        }

        if (image.getType() != BufferedImage.TYPE_BYTE_GRAY || forceNewImage) {
            BufferedImage im = new BufferedImage(image.getWidth(), image.getHeight(), BufferedImage.TYPE_BYTE_GRAY);
            Graphics g = im.getGraphics();
            g.drawImage(image, 0, 0, null);
            g.dispose();
            return im;
        } else {
            return image;
        }
    }

    public static BufferedImage getScaledVersion(BufferedImage im, int width, int height) {

        BufferedImage image = new BufferedImage(width, height, BufferedImage.TYPE_BYTE_GRAY);
        Graphics g = image.getGraphics();

        g.drawImage(im, 0, 0, width, height, null);
        g.dispose();

        return image;
    }

    public static BufferedImage getScaledVersion(BufferedImage image, double scale) {
        BufferedImage img = new BufferedImage((int) Math.round(image.getWidth() / scale), (int) Math.round(image.getHeight() / scale), BufferedImage.TYPE_BYTE_GRAY);
        Graphics g = img.getGraphics();
        g.drawImage(image, 0, 0, img.getWidth(), img.getHeight(), null);
        g.dispose();
        return img;
    }

    public static double[] image2array(BufferedImage image) {
        if (image != null) {
            image = image2grayscale(image, true);
            double[] array = new double[image.getWidth() * image.getHeight()];
            image.getData().getSamples(0, 0, image.getWidth(), image.getHeight(), 0, array);
            return array;
        } else {
            throw new IllegalArgumentException("Image cannot be null");
        }
    }

    public static BufferedImage array2image(double[] array, Dimension d) {
        BufferedImage im = new BufferedImage(d.width, d.height, BufferedImage.TYPE_BYTE_GRAY);
        DataBufferDouble db = new DataBufferDouble(array, d.width * d.height);
        BandedSampleModel sm = new BandedSampleModel(DataBuffer.TYPE_DOUBLE, d.width, d.height, 1);
        im.setData(Raster.createWritableRaster(sm, db, null));
        return im;
    }

    public static BufferedImage bytearray2image(byte[] array, Dimension d) {
        BufferedImage im = new BufferedImage(d.width, d.height, BufferedImage.TYPE_BYTE_GRAY);
        DataBuffer db = new DataBufferByte(array, d.width * d.height);
        BandedSampleModel sm = new BandedSampleModel(DataBuffer.TYPE_BYTE, d.width, d.height, 1);
        im.setData(Raster.createWritableRaster(sm, db, null));
        return im;
    }

    public static double[] getImagePatch(BufferedImage image, Rectangle coords, Dimension dim) {
        BufferedImage patch = image.getSubimage(coords.x, coords.y, coords.width, coords.height);
        patch = ImageUtils.getScaledVersion(patch, dim.width, dim.height);
        return image2array(patch);
    }

    public static BufferedImage loadImage(String filename) {
        return loadImage(new File(filename));
    }

    public static BufferedImage loadImage(File file) {
        try {
            return ImageIO.read(file);
        } catch (IOException ex) {
            throw new RuntimeException("Unable to load file: " + file.getName(), ex);
        }
    }

    public static void saveImage(BufferedImage image, String filename) {
        saveImage(image, new File(filename));
    }

    public static void saveImage(BufferedImage image, File file) {
        try {
            ImageIO.write(image, file.getName().substring(file.getName().length() - 3), file);
        } catch (IOException ex) {
            throw new RuntimeException("Unable to save file" + file.getName(), ex);
        }
    }

}
